/**********************************************************************
* 
* File: hw5.c
* Authors: Jim Guerber, Steve Wolfman 
* CSE 142 Spring 2001
* 
* Description: framework code for HW5 Part A
* See http://www.cs.washington.edu/142/ (navigate to the 
* homework page) for details.  
*********************************************************************/

/********** Includes *************************************************/
#include "GP142.h"
#include <stdio.h>
#include <assert.h>

/********** Empty or no-user constants *******************************/
#define NOBODY       -1 /* Index of player array that means "no such 
						   player" */

/********** Constants for drawing geometry ***************************/
#define FILL      0   /* Flag for filling drawing primitives */
#define LEFT      -60 /* X coordinate of left doors */
#define RIGHT     60  /* X coordinate of right doors */
#define LOWER     -60 /* Y coordinate of lower doors */
#define UPPER     60  /* Y coordinate of upper doors */
#define DOOR_SIZE 40  /* Width of the doors in pixels */

/********** Function prototypes **************************************/
void drawRoom(player_info players[], GP142_point locs[], int numPlayers);
void drawDoor(int region);
int  findPlayerIndex(player_info players[], int length, int playerID);
int  findMouseClickRegion(int mouseX, int mouseY);
int  isPointInPolygon(GP142_point point, GP142_point polygon[], int numVerts);
void setupRegion(int region, GP142_point * center, int * color, GP142_point verts[]);
void handleMouse(int mouseX, int mouseY, player_info * player, 
				 GP142_point * lastClick);


/**********************************************************************
* main:
*   Runs the program. 
*
*precondition:  nothing, void, darkness
*postcondition: again, nothing much, except the program has run!
*********************************************************************/
int main(void)
{
	/* Event loop management variables */
	int quit = FALSE;       /* Has user signaled quit yet? */
	int nextEvent;			/* The next event in the event loop */
	int mouseX, mouseY;     /* The mouse coordinates of an event */
	char keyPressed;        /* The character pressed for an event */
	player_info currPlayer; /* The player associated with an event */
	int door;               /* The door associated with an event */
  
	/* Player management variables */
	player_info players[MAX_PLAYERS]; /* Array of players in the room */
    int numPlayers = 0;               /* Number of players currently in
                                         the room */
	GP142_point lastClicks[MAX_PLAYERS];    
                                      /* Parallel array: last click location
									     of the player. (or start loc) */
	int playerIndex;                  /* Array index of current player */

	GP142_open();           /* Open and initialize the GP142 Graphics 
							   Window */
    GP142_logging(LOG_OFF); /* Logging is useful to students during 
							   debugging, but annoying during the demo; 
							   turn it off. */
	GP142_animate(ANI_RUN); /* Start animation */
							   
    /* This draws our room. */
	drawRoom(players, lastClicks, numPlayers);
							   
	/* Event loop */
	while (!quit)
	{
		/* Find out which event has happened. */
		nextEvent = GP142_await_event(&mouseX, &mouseY, &keyPressed, 
									   &door, &currPlayer);
		
		/* If the event could be somewhat relevant to the player array */
		if (nextEvent == GP142_MOUSE ||
			nextEvent == GP142_KBD ||
			nextEvent == GP142_PLAYER_LEAVE)
		{
			/* Snag the relevant player index */
			if (currPlayer.id != NULL_USER_ID)
				playerIndex = findPlayerIndex(players, numPlayers, 
											  currPlayer.id);
			else
				playerIndex = NOBODY;
		}
		
		/* Figure out what to do with our event */
		switch (nextEvent) 
		{
		case GP142_QUIT:
			quit = TRUE; /* set flag to terminate loop */
			break;
									   
		case GP142_MOUSE:
			if (playerIndex != NOBODY) 
			{
				handleMouse(mouseX, mouseY, &players[playerIndex], 
					&lastClicks[playerIndex]);
			}
			break;
									   
		case GP142_KBD:
			/* UNSUPPORTED IN GAMES VERSION! */
			break;
									   
		case GP142_PERIODIC:
			break;
									   
		case GP142_PLAYER_ENTER:
		/* TODO:
		 *   Add the incoming player to the player array and update
		 *   any needed variables (and parallel arrays) accordingly.  */
			break;
									   
		case GP142_PLAYER_LEAVE:
		/* TODO:
		 *   Remove the player associated with this event from the 
		 *   player array and update accordingly.  */
			break;
									   
		default:
			break;
		}
		/* Redraw the room each time through the loop to reflect changes. */
		drawRoom(players, lastClicks, numPlayers);
	}
	
	GP142_close(); /* Clean up and close graphics window */
	
	return 0;
}

/**********************************************************************
 * handleMouse:
 *   Updates the world to reflect the effects of the given click.
 *
 *      mouseX:    the x location of the click
 *      mouseY:    the y location of the click
 *      player:    pointer to the current player
 *      lastClick: pointer to the last click loc of the curr player
 *
 *precondition:  there is a valid current player and last click and
 *               all parameters reflect these events
 *postcondition: updates the room to handle the mouse click
 *********************************************************************/
void handleMouse(int mouseX, int mouseY, player_info * player, 
				 GP142_point * lastClick)
{
	/* Locate the "logical" location of the click. */
	int region = findMouseClickRegion(mouseX, mouseY);
	
	/* Any valid region should cause an exit event since we
	   only have door regions. */
	if (region != NO_DOOR) 
	{
		GP142_RequestLeave(*player, region);
	}
		
	/* Update the last click point. */
	lastClick->x = mouseX;
	lastClick->y = mouseY;
}


/**********************************************************************
 * drawRoom:
 *   Draws the room.  
 *
 *      players:    an array of all the players in the room
 *      locs:       an array of those players' "locations"
 *      numPlayers: the number of elts in each of the two arrays
 *
 *precondition:  players and locs are numPlayers long
 *postcondition: draws the room
 *********************************************************************/
void drawRoom(player_info players[], GP142_point locs[], int numPlayers)
{
	int i;

	GP142_clear();

	/* Draw the doors. */
	drawDoor(NORTH);
	drawDoor(EAST);
	drawDoor(SOUTH);
	drawDoor(WEST);
	
	/* Draw the player names. */
	for (i = 0; i < numPlayers; i++) 
	{
		GP142_printfP(BLACK, locs[i], 10, players[i].name);
	}
}

/**********************************************************************
 * drawDoor:
 *   Draws a door.  A "door" is a square, DOOR_SIZE pixels wide.  Joy!
 *
 *     region: the region number of the door
 *
 *precondition:  region is one of the defined regions
 *postcondition: a box of the appropriate color in the right place
 *********************************************************************/
void drawDoor(int region)
{
	GP142_point regionPoly[4];
	GP142_point center;
	int color;

	setupRegion(region, &center, &color, regionPoly);

	/* draw rectangle with opposite corners */
	GP142_rectangleP(color, regionPoly[0], regionPoly[2], FILL);
}

/**********************************************************************
 * findPlayerIndex:
 *   Searches the players array and returns the index of the element
 *   which contains the player identified by playerID.  
 *
 *     players:  the player array  NOTE: size must match "length" parm
 *     length:   length of the player array
 *     playerID: the unique ID number of the player in question
 *
 *precondition:  players is length big (or larger), active elts of
 *               players are the first length elts;
 *               playerID >= 0
 *postcondition: returns the index number of the player in the array
 *               identified by playerID. Returns NOBODY if no such 
 *               player is in the array 
 *********************************************************************/
int findPlayerIndex(player_info players[], int length, int playerID)
{
  /* TODO:
   *   Fill this in! */

  return NOBODY;
}


/**********************************************************************
 * isLeft:
 *   Given a point to check and line segment, returns whether the point
 *   is left of (>0), on (0), or right of (<0) the line defined by that
 *   segment. NOT A BOOLEAN
 *
 *     point:    the point to check
 *     lineA:    one endpoint of the line
 *     lineB:    other endpoint of the line
 *
 *precondition:  none (note that degenerate lines count as always on)
 *postcondition: > 0 for left, 0 for on, < 0 for right
 *********************************************************************/
int isLeft(GP142_point point, GP142_point lineA, GP142_point lineB)
{
	/* This code based on an implementation from:
       http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm */
	return ( (lineB.x - lineA.x) * (point.y - lineA.y)
		- (point.x - lineA.x) * (lineB.y - lineA.y) );
}

/**********************************************************************
 * isPointInPolygon:
 *   Given a point to check and the parameters of a polygon, returns
 *   true (1) if and only if the point is inside the polygon. NOTE:
 *   "inside" includes some edges of the polygon but not all. In
 *   particular, the lower and left edges/vertices are included.
 *
 *     point:    the point to check
 *     polygon:  an array of the points in the polygon
 *     numVerts: the number of vertices (points) in the polygon -- so,
 *               3 for a triangle, 4 for a square, 8 for an octagon.
 *
 *precondition:  numVerts is the number of elts in polygon; 
 *               polygon stores the vertices in order
 *postcondition: returns 1 if point is inside the polygon; 0 otherwise
 *********************************************************************/
int isPointInPolygon(GP142_point point, GP142_point polygon[], int numVerts)
{
	/* This code uses the "winding" test to determine whether the point
	   is in the polygon. Basically, it checks whether the polygon actually
	   goes around the point or not, but it's VERY robust. It should be 
	   able to handle any polygon (meeting the preconds) you throw at it. 
	 */
	
	/* This code based on an implementation from:
       http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm */

	int    wn = 0;    /* the winding number counter */
	int    i, next;
	
	/* loop through all edges of the polygon */
	for (i = 0; i < numVerts; i++) {        /* edge from polygon[i] to polygon[i+1] */
		next = (i+1) % numVerts;            /* the next vertex  with wraparound.    */
        if (polygon[i].y <= point.y) {      /* start y <= point.y                   */
            if (polygon[next].y > point.y)  /* an upward crossing                   */
                if (isLeft( polygon[i], polygon[next], point) > 0)  
					/* P left of edge */
                    wn++;                   /* have a valid up intersect            */
        }
        else {                              /* start y > point.y (no test needed)   */
            if (polygon[next].y <= point.y) /* a downward crossing                  */
                if (isLeft( polygon[i], polygon[next], point) < 0)  
					/* P right of edge */
                    wn--;                  /* have a valid down intersect           */
        }
    }

    return !(wn == 0);
}

/**********************************************************************
 * findMouseClickRegion:
 *   Given x and y coordinates, returns the region the click was in. 
 *   It's either a direction or NO_DOOR.  
 *
 *     x: the X coordinate of the click
 *     y: the Y coordinate of the click
 *
 *precondition:  none
 *postcondition: has 5 possible return values - NORTH, EAST, WEST, 
 *               SOUTH or NO_DOOR.  The return values are
 *               defined in GP142.h. 
 *********************************************************************/
int findMouseClickRegion(int x, int y)
{
	GP142_point south[4], north[4], east[4], west[4];
	GP142_point dummyPoint;
	int   dummyColor;

	/* The click point. */
	GP142_point click = {x, y};

	/* Arrays constructed from the region boundaries */
	setupRegion(SOUTH, &dummyPoint, &dummyColor, south);
	setupRegion(NORTH, &dummyPoint, &dummyColor, north);
	setupRegion(EAST, &dummyPoint, &dummyColor, east);
	setupRegion(WEST, &dummyPoint, &dummyColor, west);
	
	/* Find the region; assumes non-overlapping regions. */

	if (isPointInPolygon(click, south, 4)) 
	{
		return SOUTH;
	}

	if (isPointInPolygon(click, north, 4)) 
	{
		return NORTH;
	}

	if (isPointInPolygon(click, east, 4)) 
	{
		return EAST;
	}

	if (isPointInPolygon(click, west, 4)) 
	{
		return WEST;
	}

	/* Otherwise, return the no region value. */
	return NO_DOOR;
}


/**********************************************************************
 * setupRegion:
 *   Given a region number, changes the given center, color, 
 *   and array to reflect that region. 
 *
 *      region:  the region to generate
 *      center:  (output parameter) the center of the given region
 *      color:   (output parameter) the color of the given region
 *      verts:   (output parameter) the vertices of the polygon
 *               defining that region
 *
 *precondition:  region is one of the predefined regions:
 *                 NORTH, EAST, WEST, or SOUTH (currently)
 *               verts is a large enough array to contain all 
 *               vertices of the desired region.
 *postcondition: center, color, and verts reflect the chosen region
 *********************************************************************/
void setupRegion(int region, GP142_point * center, 
				 int * color, GP142_point verts[])
{
	assert(region == NORTH || region == SOUTH || 
		   region == EAST || region == WEST);

	switch (region) {
	case NORTH:
		center->x = RIGHT;
		center->y = UPPER;
		*color = RED;
		break;
	case EAST:
		center->x = RIGHT;
		center->y = LOWER;
		*color = BLUE;
		break;
	case SOUTH:
		center->x = LEFT;
		center->y = LOWER;
		*color = GREEN;
		break;
	case WEST:
		center->x = LEFT;
		center->y = UPPER;
		*color = YELLOW;
		break;
	default:
		/* This should not happen! */
		return;
	};

	verts[0].x = center->x - DOOR_SIZE / 2;
	verts[1].x = center->x + DOOR_SIZE / 2;
	verts[2].x = center->x + DOOR_SIZE / 2;
	verts[3].x = center->x - DOOR_SIZE / 2;

	verts[0].y = center->y - DOOR_SIZE / 2;
	verts[1].y = center->y - DOOR_SIZE / 2;
	verts[2].y = center->y + DOOR_SIZE / 2;
	verts[3].y = center->y + DOOR_SIZE / 2;
}
